# -*- coding: utf-8 -*-

"""
GAE_TwitOAuth: Twitter OAuth Consumer & API wrapper for Google App Engine

License: The MIT license
Copyright (c) 2009-2011 furyu-tei

#[Usage]
#==============================================================================
#   from GAE_TwitOAuth import GAE_TwitOAuth
#   
#   # *** Login
#     twitter = GAE_TwitOAuth()
#     login_url = twitter.get_login_url(callback_url=u"...")
#     self.redirect(login_url)
#     # would callback with queries('user_id', 'screen_name'), if successful in authentication
#   
#   # *** Call Twitter API with OAuth
#     twitter = GAE_TwitOAuth(screen_name=u"...") # or 'twitter = GAE_TwitOAuth(user_id=u"...")'
#     if twitter.oauth_enable():
#       (json,status_code) = twitter.<method-name>(...) # ex. (json,status_code) = twitter.update_status(status[,in_reply_to_status_id])
#   
#   # *** Logout
#     twitter = GAE_TwitOAuth(screen_name=u"...") # or 'twitter = GAE_TwitOAuth(user_id=u"...")'
#     logout_url = twitter.get_logout_url(callback_url=u"...")
#     self.redirect(logout_url)
#   
#   (*) need configuration file(gae_twitoauth.yaml)
#   (*) modify handlers of app.yaml as follows(default)
#       - url: /oa/.*
#         script: gae_twitoauth.py
#   
#==============================================================================
"""

__author__ = 'furyutei@gmail.com'
__version__ = '0.0.1c6'


import os,sys,cgi,re,urllib,urlparse
import time,datetime
import random,hashlib,hmac,base64
import logging
import wsgiref.handlers

from django.utils import simplejson
from google.appengine.ext import db
from google.appengine.ext import webapp
from google.appengine.api import urlfetch
from google.appengine.api.urlfetch import InvalidURLError,DownloadError,ResponseTooLargeError


#{ // optional parameters

CONFIG_FILE = u'gae_twitoauth.yaml'

(OAuth_Consumer,OAuth_ConsumerUrls,ApplicationName) = (None,None,None) # load from CONFIG_FILE

#} // end of optional parameters


#{ // general parameters

DEBUG_FLAG = True           # for webapp.WSGIApplication()
DEBUG_LEVEL = logging.DEBUG # for logger
DEBUG=True                  # False:  disable log() (wrapper of logging.debug())

TWITTER_BASE = u'http://twitter.com'
TWITTER_BASE_API = u'http://api.twitter.com/1'
TWITTER_BASE_SEARCH = u'http://search.twitter.com'

REQUEST_TOKEN_LIMIT_HOURS=1

FETCH_RETRY = 2
FETCH_DEADLINE_SECONDS = 10

DB_FETCH_LIMIT = 100 # max fetch number of datastore

MAX_STATUS_LENGTH = 140

OAuth_ServerUrls = {
  'RequestToken': TWITTER_BASE+u'/oauth/request_token',
  'AccessToken' : TWITTER_BASE+u'/oauth/access_token',
  'Authorize'   : TWITTER_BASE+u'/oauth/authorize',
}

#} // end of general parameters


#{ // datastore definitions
class dbOAuthRequestToken(db.Expando):
  token = db.StringProperty()
  secret = db.StringProperty()
  callback = db.StringProperty()
  date = db.DateTimeProperty(auto_now_add=True)

class dbOAuthAccessToken(db.Expando):
  token = db.StringProperty()
  secret = db.StringProperty()
  screen_name = db.StringProperty()
  auth_key = db.StringProperty()
  date = db.DateTimeProperty(auto_now_add=True)

#} // end of datastore definitions


#{ // log()
def log(*args):
  if not DEBUG: return
  for arg in args:
    try:
      logging.debug(arg)
    except Exception, s:
      try:
        logging.error(u'*** Error in log(): %s' % (unicode(s)))
      except:
        pass
#} // end of log()


#{ // logerr()
def logerr(*args):
  for arg in args:
    try:
      logging.error(arg)
    except Exception, s:
      try:
        logging.error(u'*** Error in logerr(): %s' % (unicode(s)))
      except:
        pass
#} // end of logerr()


#{ // def init_config()
def init_config():
  import yaml
  global init_config
  global OAuth_Consumer,OAuth_ConsumerUrls
  global ApplicationName
  
  OAuth_Consumer={}
  OAuth_ConsumerUrls={}
  
  ApplicationName = u''
  
  flg = False
  while True:
    try:
      conf_file = os.path.join(os.path.dirname(__file__), CONFIG_FILE)
      conf = yaml.load(open(conf_file).read().decode('utf-8'))
    except Exception, s:
      logerr(u'OAuth: init_config()',s)
      break
    
    mandatory_keys = ['Key','Secret',]
    counter = 0
    for _key in mandatory_keys:
      _fkey = u'Consumer'+_key
      OAuth_Consumer[_key] = conf.get(_fkey)
      if not OAuth_Consumer[_key]:
        logerr(u'OAuth: init_config(): %s not found' % _fkey)
        break
      counter += 1
    if counter!=len(mandatory_keys): break
    
    mandatory_path_names = ['Login','Callback','Logout',]
    counter = 0
    paths = conf.get('ConsumerPaths',{})
    for _key in mandatory_path_names:
      OAuth_ConsumerUrls[_key] = paths.get(_key)
      if not OAuth_ConsumerUrls[_key]:
        logerr(u'OAuth: init_config(): mandatory path "%s" not found' % (_key))
        break
      counter += 1
    if counter != len(mandatory_path_names): break
    
    ApplicationName = unicode(conf.get('ApplicationName',u''))
    
    init_config = lambda: True
    
    flg = True
    break
  return flg

#} // end of def init_config()


#{ // def cleanup_RequestToken()
def cleanup_RequestToken():
  threshold = datetime.datetime.utcnow()-datetime.timedelta(hours=REQUEST_TOKEN_LIMIT_HOURS)
  db_req_token = dbOAuthRequestToken.all().filter('date <= ', threshold)
  while True:
    token_list = db_req_token.fetch(DB_FETCH_LIMIT)
    len_token = len(token_list)
    if len_token <=0:
      break
    try:
      db.delete(token_list)
    except Exception, s:
      logerr(u'OAuth: cleanup_RequestToken()', s)
      break
    if len_token<DB_FETCH_LIMIT:
      break

#} // end of def cleanup_RequestToken()


#{ // def cleanup_AccessToken()
def cleanup_AccessToken(screen_name):
  try:
    db_acc_token = dbOAuthAccessToken.all().filter('screen_name = ',screen_name).order('-date')
    token_list = db_acc_token.fetch(DB_FETCH_LIMIT,offset=1)
  except Exception, s:
    logerr(u'OAuth: cleanup_AccessToken',s)
    db_acc_list = dbOAuthAccessToken.all().filter('screen_name = ',screen_name).fetch(DB_FETCH_LIMIT)
    db_acc_list.sort(lambda x,y: cmp(y.date,x.date))
    token_list = db_acc_list[:-1]
    
  if 0<len(token_list):
    try:
      db.delete(token_list)
    except Exception, s:
      logerr(u'OAuth: cleanup_AccessToken()', s)

#} // end of def cleanup_AccessToken()


#{ // def quote_param()
def quote_param(param):
  if not isinstance(param, basestring):
    param=unicode(param)
  return urllib.quote(param.encode('utf-8'),safe='~')

#} // end of quote_param()


#{ // def get_query_str()
def get_query_str(params):
  spairs = []
  for _k,_v in params.items():
    spairs.append(u'%s=%s' % (quote_param(_k),quote_param(_v)))
  return u'&'.join(spairs)
  
#} // end of def get_query_str()


#{ // def add_query_str()
def add_query_str(base,params):
  #query_str = urllib.urlencode(params)
  query_str = get_query_str(params)
  if re.search(u'\?',base):
    return u'%s&%s' % (base,query_str)
  else:
    return u'%s?%s' % (base,query_str)
#} // end of add_query_str()


#{ // def parse_rsp_params()
def parse_rsp_params(body):
  pairs = {}
  """
  #for str in body.split('&'):
  #  (key,val) = str.split('=')
  #  pairs[urllib.unquote(key).decode('utf8')] = urllib.unquote(val).decode('utf8')
  """
  qs = cgi.parse_qs(body)
  for _k,_vl in qs.items():
    pairs[_k] = _vl[0].decode('utf-8')
  
  return pairs

#} // end of def parse_rsp_params()


#{ // def fetch_entry()
def fetch_entry(url,method='GET',headers=None,params=None,payload=None,retry=True):
  if not headers: headers={}
  if not headers.has_key('Cache-Control'): headers['Cache-Control']='no-cache,max-age=0'
  if not headers.get('Accept-Charset'): headers['Accept-Charset']='Shift_JIS,utf-8;q=0.7,*;q=0.7'
  if not headers.get('Accept-Language'): headers['Accept-Language']='ja,en-us;q=0.7,en;q=0.3'
  
  if not payload and params:
    spairs=[]
    for key in sorted(params.keys()):
      spairs.append(u'%s=%s' % (quote_param(key),quote_param(params[key])))
    payload = u'&'.join(spairs)
  
  if payload:
    if method=='POST':
      if not headers.has_key('Content-Type'): headers['Content-Type'] = 'application/x-www-form-urlencoded'
    else:
      url = u'%s?%s' % (url,payload)
      payload = None
  
  if payload:
    try:
      payload = str(payload)
      # for error in local development environment(SDK 1.4.3): escape_encode() argument 1 must be str, not unicode
    except:
      pass
  
  entry=None
  try_num = 1+FETCH_RETRY if retry else 1
  for ci in range(try_num):
    try:
      entry = urlfetch.fetch(
        url = url,
        method = method,
        headers = headers,
        payload = payload,
        allow_truncated = False,
        follow_redirects = True,
        deadline = FETCH_DEADLINE_SECONDS,
      )
      if 200<=entry.status_code<300:
        break
    except Exception, s:
      logerr(u'Error in fetch_entry(): %s' % (url))
      if payload:
        logerr(u'payload: %s' % (payload))
      logerr(s)
      break
  
  return entry

#} // end of def fetch_entry()


#{ // def fetch_rpc()
def fetch_rpc(url,method=u'GET',headers=None,params=None,payload=None,keyid=None,callback=None):
  def handle_result(rpc):
    try:
      result=rpc.get_result()
    except InvalidURLError, s:
      callback(keyid,u'InvalidURLError')
      return
    except DownloadError, s:
      callback(keyid,u'DownloadError')
      return
    except ResponseTooLargeError, s:
      callback(keyid,u'ResponseTooLargeError')
      return
    except Exception, s:
      callback(keyid,s)
      return
    callback(keyid,result)
  
  if not headers: headers={}
  if not headers.has_key('Cache-Control'): headers['Cache-Control']='no-cache,max-age=0'
  
  if not payload and params:
    pairs=[]
    for key in sorted(params.keys()):
      pairs.append(u'%s=%s' % (quote_param(key),quote_param(params[key])))
    payload = u'&'.join(pairs)
  
  if payload:
    if method=='POST':
      headers['Content-Type'] = 'application/x-www-form-urlencoded'
    else:
      url = u'%s?%s' % (url,payload)
      payload = None
  
  rpc = urlfetch.create_rpc(deadline=10)
  if callback:
    rpc.callback = lambda:handle_result(rpc)
  
  urlfetch.make_fetch_call(rpc=rpc, url=url, method=method, headers=headers, payload=payload)
  
  return rpc
#} // end of def fetch_rpc()


#{ // def calcPayload()
def calcPayload(method=u'GET',url=u'',params={},acc_secret=None):
  pairs=[]
  for key in sorted(params.keys()):
    pairs.append(u'%s=%s' % (quote_param(key),quote_param(params[key])))
  
  payload = u'&'.join(pairs)
  
  #log(u'Signature Base String:')
  
  text = u'%s&%s&%s' % (method.upper(),quote_param(re.sub(u':\d+/',r'/',re.sub(u'[?#].*$',r'',url)).lower()),quote_param(payload))
  #log(u'  text: %s' % text)
  
  key = '%s&' % (quote_param(OAuth_Consumer['Secret']))
  if acc_secret:
    key += quote_param(acc_secret)
  #log(u'  key : %s' % key)
  
  params['oauth_signature'] = base64.b64encode(hmac.new(key, text, hashlib.sha1).digest())
  payload = u'%s&oauth_signature=%s' % (payload,quote_param(params['oauth_signature']))
  #log(u'oauth_signature : %s' % params['oauth_signature'])
  
  return (payload, params)

#} // end of calcPayload()


#{ // def createOAuthPayload()
def createOAuthPayload(method=u'GET',url=u'',opt_params={}):
  params = {
    'oauth_consumer_key': OAuth_Consumer['Key'],
    'oauth_signature_method': 'HMAC-SHA1',
    'oauth_timestamp': int(time.time()),
    'oauth_nonce': hashlib.sha1(str(random.getrandbits(64))).hexdigest(),
    'oauth_version':'1.0', # OPTIONAL
  }
  params.update(opt_params)
  return calcPayload(method=method,url=url,params=params,acc_secret=None)

#} // end of def createOAuthPayload()


#{ // def get_hash_rand_hexdigest()
def get_hash_rand_hexdigest(pre_str=None):
  if not pre_str: pre_str = u''
  ustr = u'%s-%16x' % (pre_str,random.getrandbits(64))
  hexdigest = hashlib.sha224(ustr.encode('utf-8')).hexdigest()
  return hexdigest

#} // end of get_hash_rand_hexdigest()


#{ // def get_acc_token()
def get_acc_token(user_id=None,screen_name=None,flg_create=False,kwds={}):
  db_act = None
  while True:
    if user_id:
      user_id = ApplicationName + unicode(user_id)
      _key = db.Key.from_path('dbOAuthAccessToken',user_id)
      db_act = db.get(_key)
      if not flg_create:
        break
      if db_act: # update parameters
        db_act.token = kwds.get('oauth_token')
        db_act.secret = kwds.get('oauth_token_secret')
        db_act.screen_name = kwds.get('screen_name')
        db_act.put()
        break
      def _create(user_id,kwds):
        db_act = dbOAuthAccessToken(
          key_name = user_id,
          token = kwds.get('oauth_token'),
          secret = kwds.get('oauth_token_secret'),
          screen_name = kwds.get('screen_name'),
          auth_key = get_hash_rand_hexdigest(user_id),
        )
        db_act.put()
        return db_act
      try:
        db_act = db.run_in_transaction(_create,user_id,kwds)
      except Exception, s:
        logerr(u'OAuth: Error in get_acc_token()',s)
        db_act = db.get(_key)
      break
    
    if screen_name:
      #cleanup_AccessToken(screen_name)
      #db_act = dbOAuthAccessToken.all().order('-date').filter('screen_name = ', screen_name).get()
      #if db_act and not re.search(u'^'+ApplicationName,unicode(db_act.key().name())):
      #  db_act=None
      db_acc_list = dbOAuthAccessToken.all().order('-date').filter('screen_name = ',screen_name).fetch(DB_FETCH_LIMIT)
      for _db_act in db_acc_list:
        if re.search(u'^'+ApplicationName,unicode(_db_act.key().name())):
          db_act = _db_act
          break
      break
    
    break
  
  return db_act
  
#} // end of def get_acc_token()


#{ // class GAE_TwitOAuth()
class GAE_TwitOAuth(object):
  def __init__(self,user_id=None,screen_name=None):
    init_config()
    self.db_act = db_act = get_acc_token(user_id=user_id,screen_name=screen_name)
    if db_act:
      if not user_id:
        user_id = re.sub(u'^'+ApplicationName,r'',unicode(db_act.key().name()))
      self.user_id = user_id
      self.screen_name = screen_name if screen_name else db_act.screen_name
      self.token = db_act.token
      self.secret = db_act.secret
      self.auth_key = db_act.auth_key
    else:
      self.user_id = user_id if user_id else u''
      self.screen_name = screen_name if screen_name else u''
      self.token = None
      self.secret = None
      self.auth_key = u''
  
  def oauth_enable(self):
    if self.db_act:
      return True
    else:
      return False
  
  def _get_logio_url(self,kind,callback_url):
    base_url = OAuth_ConsumerUrls[kind]
    params={}
    if kind == u'Logout':
      if self.user_id:
        params['user_id'] = self.user_id
      if self.screen_name:
        params['screen_name'] = self.screen_name
      params['auth_key'] = self.auth_key
    if callback_url:
      params['cb'] = callback_url
    url = add_query_str(base_url,params)
    return url
  
  def get_login_url(self,callback_url=None):
    return self._get_logio_url(u'Login',callback_url)
  
  def get_logout_url(self,callback_url=None):
    return self._get_logio_url(u'Logout',callback_url)
  
  def del_access_token(self):
    if self.oauth_enable():
      db.delete(self.db_act)
    self.db_act = None
    self.token = None
    self.secret = None
    self.auth_key = u''
  
  def get_user_id(self):
    return self.user_id

  def get_screen_name(self):
    return self.screen_name
  
  def _create_payload(self,method=u'GET',url=u'',opt_params={}):
    params = {
      'oauth_consumer_key': OAuth_Consumer['Key'],
      'oauth_signature_method': 'HMAC-SHA1',
      'oauth_timestamp': int(time.time()),
      'oauth_nonce': hashlib.sha1(str(random.getrandbits(64))).hexdigest(),
      'oauth_version':'1.0', # OPTIONAL
      'oauth_token':self.token,
    }
    params.update(opt_params)
    return calcPayload(method=method,url=url,params=params,acc_secret=self.secret)
  
  def _get(self,path,params):
    url = TWITTER_BASE_API + path
    (payload,params) = self._create_payload(method=u'GET',url=url,opt_params=params)
    result = fetch_entry(url=url,method='GET',payload=payload)
    
    try:
      log(u'_get(): result.headers')
      for (_key, _value) in result.headers.items():
        if re.search(u'^x-ratelimit-',_key):
          log(u'%s=%s' % (_key,_value))
    except:
      pass
    
    return result
  
  def _post(self,path,params):
    url = TWITTER_BASE_API + path
    (payload,params) = self._create_payload(method=u'POST',url=url,opt_params=params)
    result = fetch_entry(url=url,method='POST',payload=payload)
    return result
  
  def _put(self,path,params):
    url = TWITTER_BASE_API + path
    (payload,params) = self._create_payload(method=u'PUT',url=url,opt_params=params)
    result = fetch_entry(url=url,method='GET',payload=payload)
    return result
  
  def _delete(self,path,params):
    url = TWITTER_BASE_API + path
    (payload,params) = self._create_payload(method=u'DELETE',url=url,opt_params=params)
    result = fetch_entry(url=url,method='GET',payload=payload)
    return result
  
  def _rpc_get(self,path,params,keyid=None,callback=None):
    url = TWITTER_BASE_API + path
    (payload,params) = self._create_payload(method=u'GET',url=url,opt_params=params)
    rpc = fetch_rpc(url=url,method='GET',payload=payload,keyid=keyid,callback=callback)
    return rpc
  
  def _rpc_post(self,path,params,keyid=None,callback=None):
    url = TWITTER_BASE_API + path
    (payload,params) = self._create_payload(method=u'POST',url=url,opt_params=params)
    rpc = fetch_rpc(url=url,method='POST',payload=payload,keyid=keyid,callback=callback)
    return rpc
  
  def _rpc_put(self,path,params,keyid=None,callback=None):
    url = TWITTER_BASE_API + path
    (payload,params) = self._create_payload(method=u'PUT',url=url,opt_params=params)
    rpc = fetch_rpc(url=url,method='POST',payload=payload,keyid=keyid,callback=callback)
    return rpc
  
  def _rpc_delete(self,path,params,keyid=None,callback=None):
    url = TWITTER_BASE_API + path
    (payload,params) = self._create_payload(method=u'DELETE',url=url,opt_params=params)
    rpc = fetch_rpc(url=url,method='POST',payload=payload,keyid=keyid,callback=callback)
    return rpc
  
  def _sget(self,path,params):
    url = TWITTER_BASE_SEARCH + path
    (payload,params) = self._create_payload(method=u'GET',url=url,opt_params=params)
    result = fetch_entry(url=url,method='GET',payload=payload)
    
    try:
      log(u'_sget(): result.headers')
      for (_key, _value) in result.headers.items():
        if re.search(u'^x-ratelimit-',_key):
          log(u'%s=%s' % (_key,_value))
    except:
      pass
    
    return result
  
  def _json_to_dic(self,result):
    if 200<=result.status_code<300:
      try:
        dic = simplejson.loads(result.content)
      except:
        dic = {}
    else:
      dic = {}
    #log(u'### Twitter API result', u'*** status_code: %d' % result.status_code, u'*** json(dictionary):', dic)
    log(u'### Twitter API result', u'*** status_code: %d' % result.status_code)
    
    return (dic,result.status_code)
  
  # === タイムライン関連のAPI
  def public_timeline(self,since_id=None):
    path = '/statuses/public_timeline.json'
    
    params={}
    if since_id: params['since_id']=since_id
    
    return self._json_to_dic(self._get(path,params))
    
  
  def friends_timeline(self,id=None,since=None,since_id=None,max_id=None,count=None,page=None):
    path = '/statuses/friends_timeline.json'
    
    params={}
    if id: params['id']=id
    if since: params['since']=since
    if since_id: params['since_id']=since_id
    if max_id: params['max_id']=max_id
    if count: params['count']=count
    if page: params['page']=page
    
    return self._json_to_dic(self._get(path,params))
    
  
  def user_timeline(self,id=None,user_id=None,screen_name=None,count=None,since=None,since_id=None,max_id=None,page=None):
    path = '/statuses/user_timeline.json'
    
    params={}
    if id: params['id']=id
    if user_id: params['user_id']=user_id
    if screen_name: params['screen_name']=screen_name
    if count: params['count']=count
    if since: params['since']=since
    if since_id: params['since_id']=since_id
    if max_id: params['max_id']=max_id
    if page: params['page']=page
    
    return self._json_to_dic(self._get(path,params))
    
  
  def replies(self,page=None,count=None,since=None,since_id=None):
    path = '/statuses/replies.json'
    
    params={}
    if page: params['page']=page
    if count: params['count']=count
    if since: params['since']=since
    if since_id: params['since_id']=since_id
    
    return self._json_to_dic(self._get(path,params))


  def mentions(self,since_id=None,max_id=None,page=None,count=None):
    path = '/statuses/mentions.json'
    
    params={}
    if since_id: params['since_id']=since_id
    if max_id: params['max_id']=max_id
    if page: params['page']=page
    if count: params['count']=count
    
    return self._json_to_dic(self._get(path,params))
  
  
  # === ステータス関連のAPI
  def show_status(self,id):
    path = '/statuses/show/%s.json' % str(id)
    
    params={}
    
    return self._json_to_dic(self._get(path,params))
  
  
  def update_status(self,status,in_reply_to_status_id=None,source=None):
    path = '/statuses/update.json'
    
    params={'status':status[:MAX_STATUS_LENGTH]}
    if in_reply_to_status_id: params['in_reply_to_status_id']=in_reply_to_status_id
    if source: params['source']=source
    
    return self._json_to_dic(self._post(path,params))
  
  
  def update_status_rpc(self,status,in_reply_to_status_id=None,source=None,keyid=None,callback=None):
    path = '/statuses/update.json'
    
    params={'status':status[:MAX_STATUS_LENGTH]}
    if in_reply_to_status_id: params['in_reply_to_status_id']=in_reply_to_status_id
    if source: params['source']=source
    
    return self._rpc_post(path,params,keyid,callback)
  
  
  def destroy_status(self,id):
    path = '/statuses/destroy/%s.json' % str(id)
    
    params={}
    
    return self._json_to_dic(self._post(path,params))
  
  # === ユーザ情報関連のAPI
  def friends(self,id=None,user_id=None,screen_name=None,page=None,lite=None,since=None):
    path = '/statuses/friends.json'
    
    params={}
    if id: params['id']=id
    if user_id: params['user_id']=user_id
    if screen_name: params['screen_name']=screen_name
    if lite: params['lite']=lite
    if since: params['since']=since
    
    return self._json_to_dic(self._get(path,params))
  
  
  def followers(self,id=None,user_id=None,screen_name=None,page=None,lite=None):
    path = '/statuses/followers.json'
    
    params={}
    if id: params['id']=id
    if user_id: params['user_id']=user_id
    if screen_name: params['screen_name']=screen_name
    if lite: params['lite']=lite
    
    return self._json_to_dic(self._get(path,params))
  
  
  def featured(self):
    path = '/statuses/featured.json'
    
    params={}
    
    return self._json_to_dic(self._get(path,params))
  

  def show(self,id=None,user_id=None,screen_name=None):
    params={}
    if id:
      path = '/users/show/%s.json' % str(id)
    else:
      path = '/users/show.json'
      if user_id: params['user_id']=user_id
      if screen_name: params['screen_name']=screen_name
    
    return self._json_to_dic(self._get(path,params))
  
  # === ダイレクトメッセージ関連のAPI
  def direct_messages(self,since=None,since_id=None,count=None,page=None):
    path = '/direct_messages.json'
    
    params={}
    if since: params['since']=since
    if since_id: params['since_id']=since_id
    if count: params['count']=count
    if page: params['page']=page
    
    return self._json_to_dic(self._get(path,params))
  
  def sent_messages(self,since=None,since_id=None,page=None):
    path = '/direct_messages/sent.json'
    
    params={}
    if since: params['since']=since
    if since_id: params['since_id']=since_id
    if page: params['page']=page
    
    return self._json_to_dic(self._get(path,params))
    
  def new_message(self,user,text):
    path = '/direct_messages/new.json'
    
    params={'user':user,'text':text[:MAX_STATUS_LENGTH]}
    
    return self._json_to_dic(self._post(path,params))
    
  def destroy_message(self,id):
    path = '/direct_messages/destroy/%s.json' % str(id)
    
    params={}
    
    return self._json_to_dic(self._post(path,params))
  
  
  # === フレンド関連のAPI
  def create_friend(self,id,follow=None):
    path = '/friendships/create/%s.json' % str(id)
    
    params={}
    if follow: params['follow']=follow
    
    return self._json_to_dic(self._post(path,params))
  
  def destroy_friend(self,id):
    path = '/friendships/destroy/%s.json' % str(id)
    
    params={}
    
    return self._json_to_dic(self._post(path,params))
  
  def exists_friend(self,user_a,user_b):
    path = '/friendships/exists.json'
    
    params={'user_a':user_a,'user_b':user_b}
    
    return self._json_to_dic(self._get(path,params))
    
  
  # === ソーシャルグラフ関連のAPI
  def ids_friends(self,id=None,user_id=None,screen_name=None):
    path = '/friends/ids.json'
    
    params={}
    if id: params['id']=id
    if user_id: params['user_id']=user_id
    if screen_name: params['screen_name']=screen_name
    
    return self._json_to_dic(self._get(path,params))


  def ids_followers(self,id=None,user_id=None,screen_name=None):
    path = '/followers/ids.json'
    
    params={}
    if id: params['id']=id
    if user_id: params['user_id']=user_id
    if screen_name: params['screen_name']=screen_name
    
    return self._json_to_dic(self._get(path,params))


  # === アカウント関連のAPI
  def verify_credentials(self):
    path = '/account/verify_credentials.json'
    
    params={}
    
    return self._json_to_dic(self._get(path,params))
  
  
  def end_sesscion(self):
    path = '/account/end_session.json'
    
    params={}
    
    return self._json_to_dic(self._post(path,params))
  
  
  def archive(self,page=None,since=None,since_id=None):
    path = '/account/archive.json'
    
    params={}
    if page: params['page']=page
    if since: params['since']=since
    if since_id: params['since_id']=since_id
    
    return self._json_to_dic(self._post(path,params))
  
  
  def update_location(self,location):
    path = '/account/update_location.json'
    
    params={'location':location}
    
    return self._json_to_dic(self._post(path,params))
  
  
  def update_location(self,device):
    path = '/account/update_delivery_device.json'
    
    params={'device':device}
    
    return self._json_to_dic(self._post(path,params))
  
  
  def update_profile_colors(self,profile_background_color=None,profile_text_color=None,profile_link_color=None,profile_sidebar_fill_color=None,profile_sidebar_border_color=None):
    path = '/account/update_profile_colors.json'
    
    params={}
    if profile_background_color: params['profile_background_color']=profile_background_color
    if profile_text_color: params['profile_text_color']=profile_text_color
    if profile_link_color: params['profile_link_color']=profile_link_color
    if profile_sidebar_fill_color: params['profile_sidebar_fill_color']=profile_sidebar_fill_color
    if profile_sidebar_border_color: params['profile_sidebar_border_color']=profile_sidebar_border_color
    
    return self._json_to_dic(self._post(path,params))
  
  
  def update_profile_image(self,image):
    path = '/account/update_profile_image.json'
    
    params={'image':image}
    
    return self._json_to_dic(self._post(path,params))
  
  
  def update_profile_background_image(self,image):
    path = '/account/update_profile_background_image.json'
    
    params={'image':image}
    
    return self._json_to_dic(self._post(path,params))
  
  
  def rate_limit_status(self):
    path = '/account/rate_limit_status.json'
    
    params={}
    
    return self._json_to_dic(self._get(path,params))
  
  
  def update_profile(self,name=None,email=None,url=None,location=None,description=None):
    path = '/account/update_profile.json'
    
    params={}
    if name: params['name']=name
    if email: params['email']=email
    if url: params['url']=url
    if location: params['location']=location
    if description: params['description']=description
    
    return self._json_to_dic(self._post(path,params))
  
  
  # === お気に入り関連のAPI
  def favorites(self,id=None,page=None):
    path = '/favorites.json'
    
    params={}
    if id: params['id']=id
    if page: params['page']=page
    
    return self._json_to_dic(self._get(path,params))
  
  
  def create_favorite(self,id):
    path = '/favorites/create/%s.json' % str(id)
    
    params={}
    
    return self._json_to_dic(self._post(path,params))
  
  
  def destroy_favorite(self,id):
    path = '/favorites/destroy/%s.json' % str(id)
    
    params={}
    
    return self._json_to_dic(self._post(path,params))
  
  
  # === IM関連のAPI
  def follow_im(self,id):
    path = '/notifications/follow/%s.json' % str(id)
    
    params={}
    
    return self._json_to_dic(self._post(path,params))
  
  
  def leave_im(self,id):
    path = '/notifications/leave/%s.json' % str(id)
    
    params={}
    
    return self._json_to_dic(self._post(path,params))
  
  
  # === ブロック関連のAPI
  def create_block(self,id):
    path = '/blocks/create/%s.json' % str(id)
    
    params={}
    
    return self._json_to_dic(self._post(path,params))
  
  
  def destroy_block(self,id):
    path = '/blocks/destroy/%s.json' % str(id)
    
    params={}
    
    return self._json_to_dic(self._post(path,params))
  
  
  # === 補助API
  def test(self):
    path = '/help/test.json'
    
    params={}
    
    return self._json_to_dic(self._get(path,params))
  
  
  def downtime_schedule(self):
    path = '/help/downtime_schedule.json'
    
    params={}
    
    return self._json_to_dic(self._get(path,params))
  
  
  # === 検索関連のAPI
  def search(self,q,callback=None,lang=None,rpp=None,page=None,since_id=None,geocode=None,show_user=None,):
    path = '/search.json'
    
    params={'q':q}
    if callback: params['callback']=callback
    if lang: params['lang']=lang
    if rpp: params['rpp']=rpp
    if page: params['page']=page
    if since_id: params['since_id']=since_id
    if geocode: params['geocode']=geocode
    if show_user: params['show_user']=show_user
    
    return self._json_to_dic(self._get(path,params))
  
  
  def search_old(self,q,callback=None,lang=None,rpp=None,page=None,since_id=None,geocode=None,show_user=None,):
    path = '/search.json'
    
    params={'q':q}
    if callback: params['callback']=callback
    if lang: params['lang']=lang
    if rpp: params['rpp']=rpp
    if page: params['page']=page
    if since_id: params['since_id']=since_id
    if geocode: params['geocode']=geocode
    if show_user: params['show_user']=show_user
    
    return self._json_to_dic(self._sget(path,params))
  
  
  def trends(self):
    path = '/trends.json'
    
    params={}
    
    return self._json_to_dic(self._get(path,params))
  
  
  def trends_old(self):
    path = '/trends.json'
    
    params={}
    
    return self._json_to_dic(self._sget(path,params))
  
  
  def trends_current(self,exclude=None):
    path = '/trends/current.json'
    
    params={}
    if exclude: params['exclude']=exclude
    
    return self._json_to_dic(self._get(path,params))
  
  
  def trends_current_old(self,exclude=None):
    path = '/trends/current.json'
    
    params={}
    if exclude: params['exclude']=exclude
    
    return self._json_to_dic(self._sget(path,params))
  
  
  def trends_daily(self,date=None,exclude=None):
    path = '/trends/daily.json'
    
    params={}
    if date: params['date']=date
    if exclude: params['exclude']=exclude
    
    return self._json_to_dic(self._get(path,params))
  
  
  def trends_daily_old(self,date=None,exclude=None):
    path = '/trends/daily.json'
    
    params={}
    if date: params['date']=date
    if exclude: params['exclude']=exclude
    
    return self._json_to_dic(self._sget(path,params))
  
  
  def trends_weekly(self,date=None,exclude=None):
    path = '/trends/weekly.json'
    
    params={}
    if date: params['date']=date
    if exclude: params['exclude']=exclude
    
    return self._json_to_dic(self._get(path,params))
  
  
  def trends_weekly_old(self,date=None,exclude=None):
    path = '/trends/weekly.json'
    
    params={}
    if date: params['date']=date
    if exclude: params['exclude']=exclude
    
    return self._json_to_dic(self._sget(path,params))
  
  
  # === リスト関連のAPI
  def create_list(self,name,mode=None):
    path = '/%s/lists.json' % (self.screen_name)
  
    params={'name': name,}
    if mode: params['mode']=mode # 'public'/'private'
    
    return self._json_to_dic(self._post(path,params))
  
  
  def update_list(self,name,mode):
    path = '/%s/lists.json' % (self.screen_name)
  
    params={'name': name,'mode': mode,}
    return self._json_to_dic(self._post(path,params))
  
  
  def list_lists(self,name,mode):
    path = '/%s/lists.json' % (self.screen_name)
  
    params={}
    return self._json_to_dic(self._get(path,params))
  
  
  def list_lists_memberships(self,name,mode):
    path = '/%s/lists/memberships.json' % (self.screen_name)
  
    params={}
    return self._json_to_dic(self._get(path,params))
  
  
  def del_list(self,list_slug):
    path = '/%s/lists/%s.json' % (self.screen_name,list_slug)
  
    params={}
    return self._json_to_dic(self._delete(path,params))
  
  
  def get_lists_statuses(self,list_slug,next_cursor=None,previous_cursor=None):
    path = '/%s/lists/%s/statuses.json' % (self.screen_name,list_slug)
  
    params={}
    if next_cursor: params['next_cursor'] = next_cursor
    if previous_cursor: params['previous_cursor'] = previous_cursor
    
    return self._json_to_dic(self._get(path,params))
  
  
  def show_list(self,list_slug):
    path = '/%s/lists/%s.json' % (self.screen_name,list_slug)
    
    params={}
    return self._json_to_dic(self._get(path,params))
  
  
  def add_member_to_list(self,list_slug,id):
    path = '/%s/%s/members.json' % (self.screen_name,list_slug)
    
    params={'id':id,}
    return self._json_to_dic(self._post(path,params))
  
  def list_members(self,list_slug):
    path = '/%s/%s/members.json' % (self.screen_name,list_slug)
    
    params={}
    return self._json_to_dic(self._get(path,params))
  
  def remove_list_member(self,list_slug,id):
    path = '/%s/%s/members.json' % (self.screen_name,list_slug)
    
    params={'id':id,}
    return self._json_to_dic(self._delete(path,params))
  
  def is_member_in_list(self,list_slug,id):
    path = '/%s/%s/members/%s.json' % (self.screen_name,list_slug,str(id))
    
    params={}
    return self._json_to_dic(self._get(path,params))
  
  def subscribe_to_list(self,list_slug):
    path = '/%s/%s/subscribers.json' % (self.screen_name,list_slug)
    
    params={}
    return self._json_to_dic(self._post(path,params))
  
  def list_members_subscribed(self,list_slug):
    path = '/%s/%s/subscribers.json' % (self.screen_name,list_slug)
    
    params={}
    return self._json_to_dic(self._get(path,params))
    
  def unsubscribe_from_list(self,list_slug):
    path = '/%s/%s/subscribers.json' % (self.screen_name,list_slug)
    
    params={}
    return self._json_to_dic(self._delete(path,params))
  
  def is_member_subscribed_to_list(self,list_slug,id):
    path = '/%s/%s/subscribers/%s.json' % (self.screen_name,list_slug,str(id))
    
    params={}
    return self._json_to_dic(self._get(path,params))
  

#} // end of class GAE_TwitOAuth()


#{ // def common_init()
def common_init(self):
  (req,rsp) = (self.request,self.response)
  (rheaders,rcookies) = (req.headers,req.cookies)
  self.currentpath=os.path.dirname(__file__)
  self.tplpath=os.path.join(self.currentpath,'template/')
  self.urlbase=re.sub(u'[?#].*$','',req.url)
  self.urlroot=re.sub(u'(https?://[^/]*).*$',r'\1',req.url)
  self.sub_domain=re.sub(u'^https?://([^./]+).*$', r'\1', self.urlroot)
  return (req,rsp,rheaders,rcookies)

#} // end of common_init()


#{ // class LoginHandler()
class LoginHandler(webapp.RequestHandler):
  def get(self):
    (req,rsp,rheaders,rcookies)=common_init(self)
    
    cleanup_RequestToken()
    
    user_id = req.get('user_id')
    screen_name = req.get('screen_name')
    callback_url = req.get('cb','/')
    
    rt_url=OAuth_ServerUrls['RequestToken']
    (payload,params) = createOAuthPayload(
      method = u'POST',
      url = rt_url,
      opt_params = {'oauth_callback':self.urlroot+OAuth_ConsumerUrls['Callback']},
    )
    #result = fetch_entry(url=rt_url,method='POST',params=params)
    result = fetch_entry(url=rt_url,method='POST',payload=payload)
    
    if result.status_code<200 or 300<=result.status_code:
      logerr('OAuth: Error in Obtaining an Unauthorized Request Token (%d)' % result.status_code)
      self.redirect(callback_url)
      return
    
    log('### OAuth LoginHandler: result.content',result.content)
    rt_dic=parse_rsp_params(result.content)
    
    dbOAuthRequestToken(
      token = rt_dic['oauth_token'],
      secret = rt_dic['oauth_token_secret'],
      callback = callback_url
    ).put()
    
    auth_url = u'%s?oauth_token=%s' % (OAuth_ServerUrls['Authorize'],quote_param(rt_dic['oauth_token']))
    #log('User Authorization URL: %s' % (auth_url))
    
    self.redirect(auth_url)

#} // end of class LoginHandler()


#{ // class CallbackHandler()
class CallbackHandler(webapp.RequestHandler):
  def get(self):
    (req,rsp,rheaders,rcookies)=common_init(self)
    
    denied_token = req.get('denied')
    if denied_token:
      db_req = dbOAuthRequestToken.all().filter('token = ', denied_token).get()
      if not db_req:
        logerr('OAuth: Error in Obtaining User Authorization (unknown Request Token:%s)' % denied_token)
        self.error(400)
        return
      logerr('OAuth: User denied in Obtaining User Authorization')
      callback_url = db_req.callback
      db.delete(db_req)
      self.redirect(callback_url)
      return
    
    oauth_token = req.get('oauth_token')
    db_req = dbOAuthRequestToken.all().filter('token = ', oauth_token).get()
    if not db_req:
      logerr('OAuth: Error in Obtaining User Authorization (unknown Request Token:%s)' % oauth_token)
      self.error(400)
      return
    
    callback_url = db_req.callback
    db.delete(db_req)
    
    oauth_verifier = req.get('oauth_verifier')
    
    at_url=OAuth_ServerUrls['AccessToken']
    (payload,params) = createOAuthPayload(
      method = u'POST',
      url = at_url,
      opt_params = {'oauth_token':oauth_token, 'oauth_verifier':oauth_verifier},
    )
    result = fetch_entry(url=at_url,method='POST',payload=payload)
    
    if result.status_code<200 or 300<=result.status_code:
      logerr('OAuth: Error in Obtaining an Access Token (%d)' % result.status_code)
      self.redirect(callback_url)
      return
    
    log('### OAuth CallbackHandler: result.content',result.content)
    at_dic=parse_rsp_params(result.content)
    
    user_id = at_dic['user_id']
    screen_name = at_dic['screen_name']
    acc_token = at_dic['oauth_token']
    acc_secret = at_dic['oauth_token_secret']
    db_act = get_acc_token(user_id=user_id,screen_name=screen_name,flg_create=True,kwds=at_dic)
    
    if db_act:
      log(u'OAuth: Login Complete (Access Token=%s, user_id=%s, screen_name=%s)' % (acc_token, user_id, screen_name))
    else:
      logerr(u'OAuth: Login Complete, but could not get object (Access Token=%s, user_id=%s, screen_name=%s)' % (acc_token, user_id, screen_name))
    
    params={
      'user_id': user_id,
      'screen_name': screen_name,
    }
    self.redirect(add_query_str(callback_url,params))

#} // end of CallbackHandler()


#{ // class LogoutHandler
class LogoutHandler(webapp.RequestHandler):
  def get(self):
    (req,rsp,rheaders,rcookies)=common_init(self)
    
    user_id = req.get('user_id')
    screen_name = req.get('screen_name')
    auth_key = req.get('auth_key')
    
    twitter = GAE_TwitOAuth(user_id=user_id,screen_name=screen_name)
    if auth_key and twitter.auth_key == auth_key:
      log('OAuth: Logout (delete access token)')
      twitter.del_access_token()
    else:
      log('OAuth: Logout (no action)')
    
    callback_url = req.get('cb','/')
    self.redirect(callback_url)

#} // end of class LogoutHandler()


#{ // def main()
def main():
  logging.getLogger().setLevel(DEBUG_LEVEL)
  if not init_config():
    logerr('OAuth: Error in main(): init failure')
    return
  application = webapp.WSGIApplication([
    (OAuth_ConsumerUrls['Login'], LoginHandler),
    (OAuth_ConsumerUrls['Callback'], CallbackHandler),
    (OAuth_ConsumerUrls['Logout'], LogoutHandler),
  ],debug=DEBUG_FLAG)
  wsgiref.handlers.CGIHandler().run(application)

#} // def main()

# === entry point
if __name__ == "__main__":
  main()

"""
#==============================================================================
# 更新履歴
#==============================================================================
2011.04.30: version 0.0.1c6
 - ローカル開発環境(SDK 1.4.3)でエラーが発生するようになっていたものを修正。
   ※fetch_entry()でpayloadがunicodeだとエラーとなっていた。
 - その他細かい修正(スペルミスなど)。

2010.06.26: version 0.0.1c5
 - get_acc_token()でscreen_name指定検索の際、誤ってdbOAuthAccessTokenを消し込んで
   しまう不具合修正。
 
2010.06.13: version 0.0.1c4
 - 検索関連のAPI：search()で、pathを間違っていたのを修正。
 - 検索関連のAPIも基本的に http://api.twitter.com/1/* でコールするように修正。

2010.06.13: version 0.0.1c3
 - get_acc_token()でscreen_name指定検索の際、アプリケーション名(ApplicationName)が
   無視される不具合修正。

2010.05.23: version 0.0.1c2
 - APIのURI変更('http://twitter.com' → 'http://api.twitter.com/1')
   ※TWITTER_BASE_API = u'http://api.twitter.com/1'

2010.01.24: version 0.0.1c1
 - typo 修正(params=>params)
 
2010.01.23: version 0.0.1c
 - GAE_TwitOAuth(screen_name=...)でオブジェクトを作ると、access tokenのdatasotre
   (dbOAuthAccessToken)が(最新1個を残して)クリアされてしまう大バグ修正。
 - GAE_TwitOAuth.update_status_rpc() をコールするとエラーになる不具合修正。
   ※ fetch_rpc() の実装もれ。

2009.11.21: version 0.0.1b
 - ログイン情報を上書き時、screen_name,secretに不正な値が設定される不具合修正。
 - 設定ファイル(gae_twitoauth.yaml)でアプリケーション名(ApplicationName)を指定
   出来るように修正。
 - GAE_TwitOAuth()の引数にscreen_nameを指定したときに動作しない不具合修正。

2009.11.20: version 0.0.1a
 - ログアウト処理の不具合修正。
 - その他、微調整。

2009.11.19: version 0.0.1
 - とりあえずソースを Web 上に公開。
   ※実は全然テストしていないです。

"""

#■ end of file
